/****************************************************************************
*    This file is part of CEDEC_GO.                                         *
*    Copyright (c) 2004-2011 Takenori Kubo, Youichiro Miyake, CESA          *
*                                                                           *
*    History:                                                               *
*    01/07/2011 Tweaked for CEDEC CHALLANGE 2011 by CESA                    *
*                                                                           *
*    CEDEC_GO is free software: you can redistribute it and/or modify       *
*    it under the terms of the GNU General Public License as published by   *
*    the Free Software Foundation, either version 3 of the License, or      *
*    any later version.                                                     *
*                                                                           *
*    CEDEC_GO is distributed in the hope that it will be useful,            *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of         *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
*    GNU General Public License for more details.                           *
*                                                                           *
*    You should have received a copy of the GNU General Public License      *
*    along with CEDEC_GO.  If not, see <http://www.gnu.org/licenses/>.      *
*****************************************************************************/


#include "stdafx.h" // ɂCN[hECu "stdafx.h"Ɉړ܂B  Youichiro Miyake //Include files are moved to "stdafx.h" // Youichiro Miyake

using namespace std;

gtp::gtp(engine* e, istream& i, ostream& o)
  :current_id(1), in(i), out(o), go_engine(e)
{
  out.rdbuf()->pubsetbuf(0, 0);
  gtp_command_map["name"] = &gtp::cmd_name;
  gtp_command_map["version"] = &gtp::cmd_version;
  gtp_command_map["protocol_version"] = &gtp::cmd_protocol_version;

  gtp_command_map["quit"] = &gtp::cmd_quit;

  gtp_command_map["boardsize"] = &gtp::cmd_boardsize;
  gtp_command_map["komi"] = &gtp::cmd_komi;
  gtp_command_map["clear_board"] = &gtp::cmd_clear_board;

  gtp_command_map["genmove"] = &gtp::cmd_genmove;
  gtp_command_map["play"] = &gtp::cmd_play;
  gtp_command_map["undo"] = &gtp::cmd_undo;

  gtp_command_map["showboard"] = &gtp::cmd_showboard;
  gtp_command_map["final_score"] = &gtp::cmd_final_score;

  gtp_command_map["known_command"] = &gtp::cmd_known_command;
  gtp_command_map["list_commands"] = &gtp::cmd_list_commands;

  gtp_command_map["time_settings"] = &gtp::cmd_time_settings;
  gtp_command_map["time_left"] = &gtp::cmd_time_left;
}

gtp::~gtp()
{}

gtp::STATUS
gtp::gtp_success(string s)
{
  out << "=";
  if (current_id > 0) {
    out << current_id ;
	cerr << current_id;
  }

  out << " " << s << "\n\n";
  cerr << " " << s << "\n\n";

  out.flush();
  return OK;
}

gtp::STATUS
gtp::gtp_failure(string s)
{
  out << "?" ;
  if (current_id > 0) {
    out << current_id ;
  }
  out << "" << s << "\n\n";
  cerr << "" << s << "\n\n";
  out.flush();
  
  return OK;
}

void
gtp::gtp_panic()
{
  out << "! panic\n\n";
  out.flush();
}

vector<string>
gtp::preprocess_line(string& line)
{
  vector<string> v;
  istringstream line_stm(line);
  string line_nocomment;
  getline(line_stm, line_nocomment, '#');
    
  istringstream line_nocomment_stm(line_nocomment);
  string token;
  while (getline(line_nocomment_stm, token, ' ')) {
    if (token != "") {
      v.push_back(token);
    }
  }
  return v;
}

void 
gtp::main_loop(void){
  string line;

  int status = OK;
  
  while (status == OK) {
    cerr << "\n start loop" << endl;
    if (in.eof()) {
    cerr << "End Of Input" << endl;
      break;
    }
    cerr << "\n start loop continue" << endl;

    if (!getline(in, line)) {
    cerr << "Can't get line from Input" << endl;
      break;
    }
    cerr << "get line from Input: \"" << line << "\"" << endl;
    if (line == "") {
      continue;
    }
    vector<string> v = preprocess_line(line);
    if (::sscanf_s(v[0].c_str(), "%d", &current_id) == 1) {
      v.erase(v.begin());
    } else {
		cerr << "current id :" << current_id << endl;
      current_id = -1;
    }
    cerr << "v.size: " << v.size() << endl;

    string& gtp_command = v[0];
    command_map_t::iterator it = gtp_command_map.find(gtp_command);
    if (it == gtp_command_map.end()) {
      gtp_failure("unknown command");
    } else {
      cerr << "gtp cmd: " << gtp_command << endl;
      status = (this->*(it->second))(v);
      cerr << "gtp cmd: " << gtp_command << 2 << endl;
    }
    if (status == FATAL) {
      cerr << "gtp_panic" << endl;
      gtp_panic();
    }
    
  }
  cerr << "exit main loop" << endl;
}

string
gtp::encode_coord(int i, int j) {
  if (i == -1 && j == -1) {
    return "PASS";
  } else if (i < 0 || i > this->boardsize
	     || j < 0 || j > this->boardsize) {
    return "??";
  } else {
    char c = 'A' + j;
    if (c >= 'I') {
      c++;
    }
    char buf[4];
    ::sprintf(buf, "%c%d", c, this->boardsize - i);
    return string(buf);
  }
}

bool
gtp::decode_coord(string& s, int& i, int& j) {
  char c;
  int  n;
  
  transform(s.begin(),s.end(), s.begin(), ::tolower);
  if (s == "pass") {
    i = j = -1;
    return true;
  }
  if (::sscanf(s.c_str(), "%c%d", &c, &n) == 2) { // sscanf_s 
    i = this->boardsize - n;
    if ( i < 0 || i >= this->boardsize ) {
       return false;
    }
    if ('a' <= c && c <= 'z') {
      j = c - 'a' + (c > 'i'? -1: 0);
      return true;
    } else if ('A' <= c && c <= 'Z') {
      j = c - 'A' + (c > 'I'? -1: 0);
      return true;
    } else {
      return false;
    }
  }
  return false;
}

COLOR
gtp::decode_color(string& s){
  COLOR c = ERROR;
  
  transform(s.begin(),s.end(), s.begin(), ::tolower);
  if (s == "b") { // Changed "Black" to "b" for GoGUI. Youichiro Miyake
    c = BLACK;
  } else if (s == "w") { // Changed "White" to "w" for GoGUI. Youichiro Miyake
    c = WHITE;
  }
  return c;
}

// GTP command
// name
gtp::STATUS
gtp::cmd_name(vector<string>& arg) {
  return gtp_success(go_engine->name());
}

// version
gtp::STATUS
gtp::cmd_version(vector<string>& arg) {
  return gtp_success(go_engine->version());
}

// protocol_version
gtp::STATUS
gtp::cmd_protocol_version(vector<string>& arg) {
  return gtp_success("2");
}

// quit
gtp::STATUS
gtp::cmd_quit(vector<string>& arg) {
  gtp_success("");
  return QUIT;
}

// boardsize
gtp::STATUS
gtp::cmd_boardsize(vector<string>& arg) {
  int size;
  if (::sscanf_s(arg[1].c_str(), "%d", &size) < 1) {
    return gtp_failure("boardsize not an integer");
  } else if (size < MIN_BOARD || size > MAX_BOARD) {
    return gtp_failure("unacceptable size");
  }

  this->boardsize = size;
  cerr << "boardsize is " << size << endl;
  go_engine->reset(boardsize, komi);

  cerr << "reset now" << endl;
 
  /*
  static int i =0;
  if(i==0) { cout<<"="; cout<<"1"; cout<<"\n\n"; return OK;}
  i++;
  */

  return gtp_success("");
}

// komi
gtp::STATUS
gtp::cmd_komi(vector<string>& arg) {
  float k;
  if (::sscanf_s(arg[1].c_str(), "%f", &k) < 1) {
    return gtp_failure("komi not a float");
  }
  this->komi = k;
  go_engine->reset(boardsize, komi);

  return gtp_success("");
}

// clear_board
gtp::STATUS
gtp::cmd_clear_board(vector<string>& arg) {
  go_engine->reset(boardsize, komi);
  return gtp_success("");
}

// genmove
gtp::STATUS
gtp::cmd_genmove(vector<string>& arg) {
  COLOR c = decode_color(arg[1]);
  if (c != BLACK && c != WHITE) {
    return gtp_failure("invalid color");
  }

  int i, j;
  go_engine->genmove(i, j, c);
  go_engine->playmove(i, j, c);

  return gtp_success(encode_coord(i, j));
}

// play
gtp::STATUS
gtp::cmd_play(vector<string>& arg) {
  COLOR c = decode_color(arg[1]);
  int i, j;
  
  if ((c != BLACK && c != WHITE) || !decode_coord(arg[2], i, j)) {
    return gtp_failure("invalid color or coordinate");
  }

  if (!go_engine->legal_move(i, j, c)) {
    return gtp_failure("illegal move");
  }

  go_engine->playmove(i, j, c);

  return gtp_success("");
}

// undo
gtp::STATUS
gtp::cmd_undo(vector<string>& arg) {
  go_engine->undo();
  return gtp_success("");
}

// known_command
gtp::STATUS
gtp::cmd_known_command(vector<string>& arg)
{
  command_map_t::iterator it = gtp_command_map.find(arg[1]);
  if (it == gtp_command_map.end()) {
    return gtp_success("false");
  }
  return gtp_success("true");
}

// list_commands
gtp::STATUS
gtp::cmd_list_commands(vector<string>& arg)
{
  command_map_t::iterator it;
  string cmd_list;
  for (it = gtp_command_map.begin(); it != gtp_command_map.end(); it++) {
    cmd_list += it->first;
	cmd_list += " "; //"\n"; GoGui Ver1.3.0 Ƃ̘ÃG[邽 "\n"" "ɕύXB Youichiro Miyake // For connection with GoGui Ver1.3.0, "\n" is changed to ""; 

  }
  //cmd_list += "\n"; //"\n"; GoGui Ve1.3.0 ƘÃG[邽߃RgAEgB Youichiro Miyake // For connection with GoGui Ver1.3.0, this line is commented-out; 

  return gtp_success(cmd_list);
}

// showboard
gtp::STATUS
gtp::cmd_showboard(vector<string>& arg)
{
  int i, j;
  string letters = "  ";
  
  for (i = 0; i < boardsize; i++) {
    letters += (" " + ('A' + i + (i >= 8? 1: 0)));
  }
  letters += "\n";

  string board;
  
  for (i = 0; i < boardsize; i++) {
    char buf[8];
    sprintf(buf, "\n%2d", boardsize - i);
    board += buf;
    for (j = 0; j < boardsize; j++) {
      switch (go_engine->get_board(i, j)) {
      case BLACK:
	board += " X";
	break;
      case WHITE:
	board += " O";
	break;
      case EMPTY:
	board += " .";
	break;
      default:
	;
      }
    }
    sprintf(buf, " %d", boardsize - i);
    board += buf;
  }
  board = letters + board + "\n" + letters;
  return gtp_success(board);
}

// final_score
gtp::STATUS
gtp::cmd_final_score(vector<string>& arg)
{
  float score = komi;
  int i, j;

  go_engine->compute_final_status();
  for (i = 0; i < boardsize; i++) {
    for (j = 0; j < boardsize; j++) {
      int status = go_engine->get_final_status(i, j);
      if (status == BLACK_TERRITORY)
	score--;
      else if (status == WHITE_TERRITORY)
	score++;
      else if ((status == ALIVE) ^ (go_engine->get_board(i, j) == WHITE))
	score--;
      else
	score++;
    }
  }

  char buf[8];
  ::sprintf(buf, "%3.1f", abs(score));
  if (score > 0.0) {
    string res = "W+";
    return gtp_success(res + buf);
  } else if (score < 0.0) {
    string res = "B+";
    return gtp_success(res + buf);
  }
  return gtp_success("0");
}

// time_settings
gtp::STATUS
gtp::cmd_time_settings(vector<string>& arg)
{
  int main_time;
  int byo_yomi_time;
  int byo_yomi_stones;

  if (::sscanf_s(arg[1].c_str(), "%d", &main_time) < 1 ||
      ::sscanf_s(arg[2].c_str(), "%d", &byo_yomi_time) < 1 ||
      ::sscanf_s(arg[3].c_str(), "%d", &byo_yomi_stones) < 1) {
    return gtp_failure("not three integers");
  }
  
  go_engine->time_settings(main_time, byo_yomi_time, byo_yomi_stones);

  return gtp_success("");
}

// time_left
gtp::STATUS
gtp::cmd_time_left(vector<string>& arg)
{
  int time;
  int stones;

  COLOR c = decode_color(arg[1]);
  if (c != BLACK && c != WHITE) {
    return gtp_failure("invalid color");
  }

  if (::sscanf_s(arg[2].c_str(), "%d", &time) < 1 ||
      ::sscanf_s(arg[3].c_str(), "%d", &stones) < 1) {
    return gtp_failure("time and stones not two integers");
  }

  go_engine->time_left(c, time, stones);
  
  return gtp_success("");
}



